"""Constants of DiffuPath."""

import os

from diffupy.constants import METHODS

dir_path = os.path.dirname(os.path.realpath(__file__))
SOURCE_DIR = os.path.join(os.path.abspath(os.path.join(dir_path, os.pardir)))

#: Default DiffuPath directory
DEFAULT_DIFFUPATH_DIR = os.path.join(os.path.expanduser('~'), '.diffupath')
#: Default DiffuPath output directory
OUTPUT_DIR = os.path.join(DEFAULT_DIFFUPATH_DIR, 'output')


def ensure_output_dirs():
    """Ensure that the output directories exists."""
    os.makedirs(DEFAULT_DIFFUPATH_DIR, exist_ok=True)
    os.makedirs(OUTPUT_DIR, exist_ok=True)


ensure_output_dirs()

"""Available diffusion cross-validation methods"""

#: raw
BY_METHOD = 'by_method'
BY_DB = 'by_db'

EVALUATION_METHODS = {
    BY_METHOD,
    BY_DB,
}

DATASETS = {
    '1',
    '2',
    '3',
}

# Rename DiffuPy methods
DIFFUPY_METHODS = METHODS

EMOJI = "🌐"

# Available databases from PathMe
#: KEGG
KEGG_NAME = 'kegg'
#: Reactome
REACTOME_NAME = 'reactome'
#: WikiPathways
WIKIPATHWAYS_NAME = 'wikipathways'

# Complementary databases from Bio2BEL
#: MirTarBase
MIRTARBASE_NAME = 'mirtarbase'
#: SIDER
SIDER_NAME = 'sider'
#: PhewasCatalog
PHEWAS_NAME = 'phewascatalog'
#: HSDN
HSDN_NAME = 'hsdn'
#: DDR
DDR_NAME = 'ddr'
#: DrugBank
DRUGBANK_NAME = 'drugbank'
#: Gene Ontology
GENE_ONTOLOGY_NAME = 'go'

#: Databases available for download in DiffuPath
DATABASES = [
    KEGG_NAME,
    REACTOME_NAME,
    WIKIPATHWAYS_NAME,
    MIRTARBASE_NAME,
    SIDER_NAME,
    PHEWAS_NAME,
    HSDN_NAME,
    DDR_NAME,
    DRUGBANK_NAME,
    GENE_ONTOLOGY_NAME,
]



# -*- coding: utf-8 -*-

"""This module has utilities methods to mine and retrieve PathMe content."""

import os
from collections import defaultdict

import pybel
from pathme.constants import KEGG_BEL, REACTOME_BEL, WIKIPATHWAYS_BEL
from pybel.constants import ANNOTATIONS
from pybel.dsl import Abundance, BiologicalProcess, CentralDogma, ListAbundance, Reaction


def calculate_database_sets_as_dict(nodes, database):
    gene_nodes, mirna_nodes, metabolite_nodes, bp_nodes = calculate_database_sets(nodes, database)
    return {'gene_nodes': gene_nodes, 'mirna_nodes': mirna_nodes, 'metabolite_nodes': metabolite_nodes,
            'bp_nodes': bp_nodes}


def get_nodes_in_database(folder):
    """Merge all python pickles in a given folder and returns the corresponding BELGraph."""
    database_networks = [
        pybel.from_pickle(os.path.join(folder, path))
        for path in os.listdir(folder)
        if path.endswith('.pickle')
    ]

    return {
        node
        for network in database_networks
        for node in network.nodes()
    }


def process_reactome_multiple_genes(genes):
    """Process a wrong ID with multiple identifiers"""
    gene_list = []

    for counter, gene in enumerate(genes):

        # Strip the ' gene' prefix
        gene = gene.strip().strip(' gene').strip(' genes')

        # First element is always OK
        if counter == 0:
            gene_list.append(gene)

        # If the identifier starts the same than the first one, it is right
        elif gene[:2] == genes[0][:2]:
            gene_list.append(gene)

        # If the identifier is longer than 2 it is a valid HGNC symbol
        elif len(gene) > 2:
            gene_list.append(gene)

        # If they start different, it might have only a number (e.g., 'ABC1, 2, 3') so it needs to be appended
        elif gene.isdigit():
            gene_list.append(genes[0][:-1] + gene)

        # If the have only one letter (e.g., HTR1A,B,D,E,F,HTR5A)
        elif len(gene) == 1:
            gene_list.append(genes[0][:-1] + gene)

    return gene_list


def munge_reactome_gene(gene):
    """Process Reactome gene"""
    if "," in gene:
        return process_reactome_multiple_genes(gene.split(","))

    elif "/" in gene:
        return process_reactome_multiple_genes(gene.split("/"))

    return gene


def calculate_database_sets(nodes, database):
    """Calculate node sets for each modality in the database"""

    # Entities in WikiPathways that required manual curation
    WIKIPATHWAYS_BIOL_PROCESS = {'lipid biosynthesis', 'hsc survival', 'glycolysis & gluconeogenesis',
                                 'triacylglyceride  synthesis', 'wnt canonical signaling',
                                 'regulation of actin skeleton', 'fatty acid metabolism',
                                 'mrna processing major splicing pathway', 'senescence', 'monocyte differentiation',
                                 'pentose phosphate pathway', 'ethanolamine  phosphate', 'hsc differentiation',
                                 'actin, stress fibers and adhesion', 'regulation of actin cytoskeleton',
                                 's-phase progression', 'g1-s transition', 'toll-like receptor signaling pathway',
                                 'regulation of  actin cytoskeleton', 'proteasome degradation', 'apoptosis',
                                 'bmp pathway', 'ampk activation', 'g1/s checkpoint arrest', 'mapk signaling pathway',
                                 'chromatin remodeling and  epigenetic modifications', 'wnt signaling pathway',
                                 'ros production', 'erbb signaling pathway', 'shh pathway', 'inflammation',
                                 'dna replication', 'mrna translation', 'oxidative stress',
                                 'cell cycle checkpoint activation', 'gi/go pathway', 'wnt pathway',
                                 'g1/s transition of mitotic cell cycle', 'modulation of estrogen receptor signalling',
                                 'dna repair', 'bmp canonical signaling', 'igf and insuline signaling',
                                 'unfolded protein response', 'cell death', 'p38/mapk  pathway', 'glycogen metabolism',
                                 'gnrh signal pathway',
                                 'the intra-s-phase checkpoint mediated arrest of cell cycle progression', 'tca cycle',
                                 'mtor protein kinase signaling pathway', 'proteasome  degradation pathway',
                                 'morphine metabolism', 'hsc aging', 'gastric pepsin release',
                                 'parietal cell production', 'prostaglandin pathway', 'cell cycle (g1/s)  progression',
                                 'notch pathway', 'g2/m progression', 'wnt signaling', 'cell adhesion',
                                 'cell cycle progression', 'egfr pathway', 'cell cycle', 'angiogenesis',
                                 'g2/m-phase checkpoint', 'hsc self renewal', '26s proteasome  degradation',
                                 'mapk signaling', 'immune system up or down regulation', 'm-phase progression',
                                 'insulin signaling', 'nf kappa b pathway', 'cell cycle  progression', 'gi pathway',
                                 'cd45+ hematopoietic-    derived cell    proliferation', "kreb's cycle",
                                 'glycogen synthesis', 'apoptosis pathway', 'g1/s progression',
                                 'inflammasome activation', 'melanin biosynthesis', 'proteasomal degradation',
                                 'g2/m checkpoint arrest', 'g1/s cell cycle transition', 'dna damage response',
                                 'gastric histamine release'}
    WIKIPATHWAYS_METAB = {'2,8-dihydroxyadenine', '8,11-dihydroxy-delta-9-thc', 'adp-ribosyl', 'cocaethylene',
                          'dhcer1p', 'ecgonidine', 'f2-isoprostane', 'fumonisins b1', 'iodine', 'l-glutamate',
                          'lactosylceramide', 'methylecgonidine', 'n-acetyl-l-aspartate', 'nad+', 'nadph oxidase',
                          'neuromelanin', 'nicotinic acid (na)', 'nmn', 'pip2', 'sphingomyelin', 'thf'}
    WIKIPATHWAYS_NAME_NORMALIZATION = {"Ca 2+": "ca 2+", "acetyl coa": "acetyl-coa", "acetyl-coa(mit)": "acetyl-coa",
                                       "h20": "h2o"}

    # Entities in Reactome that required manual curation
    BLACK_LIST_REACTOME = {"5'"}
    REACTOME_PROT = {'phospho-g2/m transition proteins', 'integrin alpha5beta1, integrin alphavbeta3, cd47',
                     'food proteins', 'activated fgfr2', 'adherens junction-associated proteins',
                     'pi3k mutants,activator:pi3k', 'prolyl 3-hydroxylases', 'gpi-anchored proteins', 'c3d, c3dg, ic3b',
                     'c4s/c6s chains', 'activated fgfr1 mutants and fusions', 'activated fgfr3 mutants', 'protein',
                     'cyclin a2:cdk2 phosphorylated g2/m transition protein', 'c4c, c3f', 'activated raf/ksr1',
                     'activated fgfr1 mutants', 'g2/m transition proteins', 'lman family receptors', 'cyclin',
                     'usp12:wdr48:wdr20,usp26', 'proteins with cleaved gpi-anchors', 'activated fgfr2 mutants',
                     'c4d, ic3b', 'c5b:c6:c7, c8, c9', 'cyclin a1:cdk2 phosphorylated g2/m transition protein',
                     'genetically or chemically inactive braf', 'il13-downregulated proteins',
                     'activated fgfr4 mutants', 'rna-binding protein in rnp (ribonucleoprotein) complexes',
                     'effector proteins', 'usp3, saga complex', 'dephosphorylated "receiver" raf/ksr1'}

    gene_nodes = set()
    mirna_nodes = set()
    metabolite_nodes = set()
    bp_nodes = set()

    for node in nodes:

        if isinstance(node, ListAbundance) or isinstance(node, Reaction) or not node.name:
            continue

        # Lower case name and strip quotes or white spaces
        name = node.name.lower().strip('"').strip()

        # Dealing with Genes/miRNAs
        if isinstance(node, CentralDogma):

            ##################
            # miRNA entities #
            ##################

            if name.startswith("mir"):

                # Reactome preprocessing to flat multiple identifiers
                if database == 'reactome':
                    reactome_cell = munge_reactome_gene(name)
                    if isinstance(reactome_cell, list):
                        for name in reactome_cell:
                            mirna_nodes.add(name.replace("mir-", "mir"))
                    else:
                        mirna_nodes.add(name.strip(' genes').replace("mir-", "mir"))

                    continue

                mirna_nodes.add(name.replace("mir-", "mir"))

            ##################
            # Genes entities #
            ##################

            else:
                # Reactome preprocessing to flat multiple identifiers
                if database == 'reactome':
                    reactome_cell = munge_reactome_gene(name)
                    if isinstance(reactome_cell, list):
                        for name in reactome_cell:
                            if name in BLACK_LIST_REACTOME:  # Filter entities in black list
                                continue
                            elif name.startswith("("):  # remove redundant parentheses
                                name = name.strip("(").strip(")")

                            gene_nodes.add(name)
                    else:
                        gene_nodes.add(name)
                    continue

                # WikiPathways and KEGG do not require any processing of genes
                if name in WIKIPATHWAYS_BIOL_PROCESS:
                    bp_nodes.add(name)
                    continue
                gene_nodes.add(name)

        #######################
        # Metabolite entities #
        #######################

        elif isinstance(node, Abundance):

            if database == 'wikipathways':
                # Biological processes that are captured as abundance in BEL since they were characterized wrong in WikiPathways
                if name in WIKIPATHWAYS_BIOL_PROCESS:
                    bp_nodes.add(name)
                    continue

                elif node.namespace in {'WIKIDATA', 'WIKIPATHWAYS', 'REACTOME'} and name not in WIKIPATHWAYS_METAB:
                    bp_nodes.add(name)
                    continue

                # Fix naming in duplicate entity
                if name in WIKIPATHWAYS_NAME_NORMALIZATION:
                    name = WIKIPATHWAYS_NAME_NORMALIZATION[name]

            elif database == 'reactome':
                # Curated proteins that were coded as metabolites
                if name in REACTOME_PROT:
                    gene_nodes.add(name)
                    continue

                # Flat multiple identifiers (this is not trivial because most of ChEBI names contain commas,
                # so a clever way to fix some of the entities is to check that all identifiers contain letters)
                elif "," in name and all(
                        string.isalpha()
                        for string in name.split(",")
                ):
                    for string in name.split(","):
                        metabolite_nodes.add(string)
                    continue

            metabolite_nodes.add(name)

        #################################
        # Biological Processes entities #
        #################################

        elif isinstance(node, BiologicalProcess):
            if name.startswith('title:'):
                name = name[6:]  # KEGG normalize

            bp_nodes.add(name)

    return gene_nodes, mirna_nodes, metabolite_nodes, bp_nodes


def get_set_database(database):
    """Return database content subsets by entity for a given db name."""

    if database == 'kegg':
        nodes = get_nodes_in_database(KEGG_BEL)

    if database == 'reactome':
        nodes = get_nodes_in_database(REACTOME_BEL)

    if database == 'wikipathways':
        nodes = get_nodes_in_database(WIKIPATHWAYS_BEL)

    return calculate_database_sets(nodes, database)


def get_labels_by_db_and_omic_from_pathme(databases):
    db_entites = defaultdict(dict)
    entites_db = defaultdict(lambda: defaultdict(set))

    for db in databases:
        genes, mirna, metabolites, bps = get_set_database(db)
        db_entites[db] = {'genes': genes, 'mirna': mirna, 'metabolites': metabolites, 'bps': bps}

        for entity_type, entities in db_entites[db].items():
            entites_db[entity_type][db] = entities

    return db_entites, entites_db


def get_labels_by_db_and_omic_from_graph(graph):
    db_subsets = defaultdict(set)
    db_entites = defaultdict(dict)
    entites_db = defaultdict(dict)

    # entity_type_map = {'Gene':'genes', 'mirna_nodes':'mirna', 'Abundance':'metabolites', 'BiologicalProcess':'bps'}

    for u, v, k in graph.edges(keys=True):
        if ANNOTATIONS not in graph[u][v][k]:
            continue

        if 'database' not in graph[u][v][k][ANNOTATIONS]:
            continue

        for database in graph[u][v][k][ANNOTATIONS]['database']:
            db_subsets[database].add(u)
            db_subsets[database].add(v)

    for database, nodes in db_subsets.items():
        db_entites[database] = calculate_database_sets_as_dict(nodes, database)

        database_sets = calculate_database_sets_as_dict(nodes, database)

        db_entites[database] = database_sets

        for entity_type, entities in database_sets.items():
            entites_db[entity_type][database] = entities

    return db_entites, entites_db



# -*- coding: utf-8 -*-

"""Miscellaneous utils of the package."""

import itertools
import logging
import pickle
import random

from statistics import mean

import numpy as np

log = logging.getLogger(__name__)


def from_pickle(input_path):
    """Read network from pickle."""
    with open(input_path, 'rb') as f:
        unpickler = pickle.Unpickler(f)
        background_mat = unpickler.load()

    return background_mat


def to_pickle(to_pickle, output):
    """Write pickle."""
    with open(output, 'wb') as file:
        pickle.dump(to_pickle, file)


def print_dict_dimensions(entities_db, title):
    """Print dimension of the dictionary"""
    total = 0

    for k1, v1 in entities_db.items():
        m = ''
        if isinstance(v1, dict):
            for k2, v2 in v1.items():
                m += f'{k2}({len(v2)}), '
                total += len(v2)
        else:
            m += f'{len(v1)} '
            total += len(v1)

        print(f'Total number of {k1}: {m} ')

    print(f'Total: {total} ')


def get_labels_set_from_dict(entities):
    if isinstance(list(entities.values())[0], dict):
        # TODO: Check
        return set(itertools.chain.from_iterable(itertools.chain.from_iterable(entities.values())))
    else:
        return set(itertools.chain.from_iterable(entities.values()))


def reduce_dict_dimension(dict):
    """Reduce dictionary dimension."""
    return {
        k: set(itertools.chain.from_iterable(entities.values()))
        for k, entities in dict.items()
    }


def check_substrings(dataset_nodes, db_nodes):
    mapping_substrings = set()

    for entity in dataset_nodes:
        if isinstance(entity, tuple):
            for subentity in entity:
                for entity_db in db_nodes:
                    if isinstance(entity_db, tuple):
                        for subentity_db in entity_db:
                            if subentity_db in subentity or subentity in subentity_db:
                                mapping_substrings.add(entity_db)
                                break
                        break
                    else:
                        if entity_db in subentity or subentity in entity_db:
                            mapping_substrings.add(entity_db)
                            break
        else:
            for entity_db in db_nodes:
                if isinstance(entity_db, tuple):
                    for subentity_db in entity_db:
                        if subentity_db in entity or entity in subentity_db:
                            mapping_substrings.add(entity_db)
                            break
                    break
                else:
                    if entity_db in entity or entity in entity_db:
                        mapping_substrings.add(entity_db)
                        break

    return mapping_substrings

def split_random_two_subsets(to_split):
    half_1 = random.sample(population=list(to_split), k=int(len(to_split) / 2))
    half_2 = list(set(to_split) - set(half_1))

    return half_1, half_2

def hide_true_positives(to_split, k=0.5):
    """Hide relative number of labels."""
    if isinstance(to_split, set):
        to_split = list(to_split)

    new_labels = to_split[:]

    # Check for -1
    if -1 in new_labels:
        new_labels = [0 if label == -1 else label for label in new_labels]

    indices = [index for index, label in enumerate(new_labels) if label != 0]

    for index in random.choices(indices, k=int(k * len(indices))):
        new_labels[index] = 0

    return new_labels


def split_random_three_subsets(to_split):
    half_1 = random.sample(population=list(to_split), k=int(len(to_split) / 3))
    half_2, half_3 = split_random_two_subsets(list(set(to_split) - set(half_1)))

    return half_1, half_2, half_3


def get_three_venn_intersections(set1, set2, set3):
    set1, set2, set3 = set(set1), set(set2), set(set3)
    set1_set2 = set1.intersection(set2)
    set1_set3 = set1.intersection(set3)
    core = set1_set3.intersection(set1_set2)

    set1_set2 = set1_set2 - core
    set1_set3 = set1_set3 - core
    set2_set3 = set2.intersection(set3) - core

    return {'unique_set1': set1 - set1_set2 - set1_set3 - core,
            'unique_set2': set2 - set1_set2 - set2_set3 - core,
            'set1_set2': set1_set2,
            'unique_set3': set3 - set1_set3 - set2_set3 - core,
            'set1_set3': set1_set3,
            'set2_set3': set2_set3,
            'core': core,
            }


def random_disjoint_intersection_two_subsets(unique_set1, unique_set2, intersection):
    set1, set2 = split_random_two_subsets(intersection)

    return unique_set1 | set(set1), unique_set2 | set(set2)


def random_disjoint_intersection_three_subsets(sets_dict):
    set_labels = list(sets_dict.keys())
    set_values = list(sets_dict.values())

    set1, set2, set3 = set_values[0][0], set_values[1][0], set_values[2][0]

    intersections = get_three_venn_intersections(set1, set2, set3)

    set1, set2 = random_disjoint_intersection_two_subsets(
        intersections['unique_set1'],
        intersections['unique_set2'],
        intersections['set1_set2']
    )

    set1, set3 = random_disjoint_intersection_two_subsets(set1,
                                                          intersections['unique_set3'],
                                                          intersections['set1_set3']
                                                          )

    set2, set3 = random_disjoint_intersection_two_subsets(set2,
                                                          set3,
                                                          intersections['set2_set3']
                                                          )

    set1_core, set2_core, set3_core = split_random_three_subsets(intersections['core'])

    return {set_labels[0]: set1 | set(set1_core),
            set_labels[1]: set2 | set(set2_core),
            set_labels[2]: set3 | set(set3_core)
            }


def get_count_and_labels_from_two_dim_dict(mapping_by_database_and_entity):
    db_labels = []
    types_labels = []

    all_count = []
    all_percentage = []

    # entity_type_map = {'metabolite_nodes': 'metabolite', 'mirna_nodes': 'micrornas', 'gene_nodes': 'genes', 'bp_nodes': 'bps'}

    for db_name, entities_by_type in mapping_by_database_and_entity.items():
        db_count = []
        db_percentage = []

        db_labels.append(db_name)

        if not types_labels:
            types_labels = entities_by_type[0].keys()

        for entity_type, entities_tupple in entities_by_type[0].items():
            db_count.append(len(entities_tupple[0]))
            db_percentage.append(entities_tupple[1])

        all_count.append(db_count)
        all_percentage.append(db_percentage)

    return np.array(all_count), np.array(all_percentage), db_labels, types_labels


def get_mean_from_two_dim_dict(d):
    for k1, v1 in d.items():
        for k2, v2 in v1.items():
            if v2:
                d[k1][k2] = [mean(v2)]

    return d



# -*- coding: utf-8 -*-

"""Command line interface."""

import json
import logging
import sys

import click
from bio2bel.constants import get_global_connection
from diffupy.diffuse import diffuse as run_diffusion
from diffupy.process_input import process_input
from diffupy.utils import process_network_from_cli

from .constants import *
from .cross_validation import cross_validation_by_method
from .input_mapping import process_input_from_cli
from .validation_datasets_parsers import parse_set1, parse_set2, parse_set3

logger = logging.getLogger(__name__)

#: Parsing methods for each dataset
PARSING_METHODS = {
    '1': parse_set1,
    '2': parse_set2,
    '3': parse_set3,
}


@click.group(help='DiffuPy')
def main():
    """Run DiffuPy."""
    logging.basicConfig(format="%(asctime)s - %(levelname)s - %(name)s - %(message)s")


@main.group()
def diffusion():
    """Commands for running diffusion experiments."""


@diffusion.command()
@click.option(
    '-n', '--network',
    help='Path to the network graph or kernel',
    required=True,
    type=click.Path(exists=True, dir_okay=False)
)
@click.option(
    '-i', '--data',
    help='Input data',
    required=True,
    type=click.Path(exists=True, dir_okay=False)
)
@click.option(
    '-o', '--output',
    type=click.File('w'),
    help="Output file",
    default=sys.stdout,
)
@click.option(
    '-m', '--method',
    help='Diffusion method',
    type=click.Choice(METHODS),
    required=True,
)
@click.option(
    '-b', '--binarize',
    help='If logFC provided in dataset, convert logFC to binary (e.g., up-regulated entities to 1, down-regulated to '
         '-1). For scoring methods that accept quantitative values (i.e., raw & z), node labels can also be codified '
         'with LogFC (in this case, set binarize==False).',
    type=bool,
    default=True,
    show_default=True,
)
@click.option(
    '-t', '--threshold',
    help='Codify node labels by applying a threshold to logFC in input.',
    type=float,
)
@click.option(
    '-a', '--absolute_value',
    help='Codify node labels by applying threshold to |logFC| in input. If absolute_value is set to False, node labels '
         'will be signed.',
    type=bool,
    default=True,
    show_default=True,
)
@click.option(
    '-p', '--p_value',
    help='Statistical significance (p-value).',
    type=float,
    default=0.05,
    show_default=True,
)
def diffuse(
    network: str,
    data: str,
    output: str,
    method: str,
    binarize: bool,
    absolute_value: bool,
    threshold: float,
    p_value: float,
):
    """Run a diffusion method over a network or pre-generated kernel."""
    click.secho(f'{EMOJI} Loading graph from {network} {EMOJI}')
    graph = process_network_from_cli(network)

    click.secho(
        f'{EMOJI} Graph loaded with: \n'
        f'{graph.number_of_nodes()} nodes\n'
        f'{graph.number_of_edges()} edges\n'
        f'{EMOJI}'
    )

    click.secho(f'Codifying data from {data}.')

    input_scores_dict = process_input(data, method, binarize, absolute_value, p_value, threshold)

    click.secho(f'Running the diffusion algorithm.')

    results = run_diffusion(
        input_scores_dict,
        method,
        graph,
    )

    json.dump(results, output, indent=2)

    click.secho(f'Finished!')


@diffusion.command()
@click.option(
    '-d', '--data',
    help='Input data',
    required=True,
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    '-n', '--network',
    help='Path to the network graph or kernel',
    required=True,
    type=click.Path(exists=True, dir_okay=False)
)
@click.option(
    '-g', '--graph_path',
    help='Path to the network as a graph',
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    '-q', '--quantitative',  # TODO Automatize if possible, check type of label_input.
    help='Generate categorical label_input from labels',
    is_flag=False,
)
@click.option(
    '-n', '--network_as_graph',
    help='If given expects graph else expects as a kernel',
    is_flag=False,
)
@click.option(
    '-o', '--output',
    help='Output path for the results',
    default=OUTPUT_DIR,
    show_default=True,
    type=click.Path(exists=True, file_okay=False),
)
@click.option(
    '-i', '--iterations',
    help='Number of distinct cross validations',
    default=25,
    show_default=True,
    type=int,
)
@click.option(
    '-c', '--comparison',
    help='Comparison method',
    default='by_method',
    show_default=True,
    type=click.Choice(EVALUATION_METHODS),
)
@click.option(
    '-k', '--dataset',
    help='Key for the datasets presented in the paper',
    show_default=True,
    default=1,
    type=click.Choice(DATASETS),
)
def evaluate(
    data: str,
    network: str,
    graph_path: str,
    quantitative: bool,  # TODO Automatize if possible, check type of label_input.
    network_as_graph: bool,  # TODO Automatize if possible, check type of graph.
    output: str,
    iterations: int,
    comparison: str,
    dataset: int,
):
    """Evaluate a kernel/network on one of the three presented datasets."""
    click.secho(f'{EMOJI} Loading label_input for cross-validation... {EMOJI}')

    if not network_as_graph and not graph_path:
        raise ValueError("Network not provided in graph format, which is required for evaluation.")

    _, kernel, labels_mapping, graph = process_input_from_cli(
        PARSING_METHODS[dataset],
        network,
        data,
        network_as_graph,
        quantitative,
    )

    if not network_as_graph:
        graph = process_network_from_cli(graph_path)

    if comparison == 'by_method':
        click.secho(f'{EMOJI} Evaluating by method... {EMOJI}')

        auroc_metrics, auprc_metrics = cross_validation_by_method(
            labels_mapping,
            graph,
            kernel,
            k=iterations,
        )
    elif comparison == 'by_db':
        click.secho(f'{EMOJI} Evaluating by database... {EMOJI}')

        # TODO to adapt from 'get_one_x_in_cv_inputs_from_subsets', and label_input treatment subset division.
        auroc_metrics, auprc_metrics = cross_validation_by_method(
            labels_mapping,
            graph,
            kernel,
            k=iterations,
        )
    else:
        raise ValueError("The comparison method provided not match any provided method.")

    with open(os.path.join(output, 'metrics.json'), 'w') as outfile:
        json.dump(
            {'auroc_metrics': auroc_metrics,
             'auprc_metrics': auprc_metrics
             },
            outfile,
            indent=2,
        )

    click.secho(f'{EMOJI} Random cross-validation performed with success. Output located at {output}... {EMOJI}')


@main.group()
def databases():
    """Commands related to available databases."""


@databases.command()
def ls():
    """Print the list of the available databases."""
    click.secho(f'{EMOJI} Available databases {DATABASES}')


@databases.command()
@click.option(
    '-d', '--database',
    help='Database',
    type=click.Choice(DATABASES),
    required=True,
)
@click.option(
    '-c',
    '--connection',
    default=get_global_connection(),
    show_default=True,
    help='Bio2BEL database connection string',
)
@click.option('-v', '--verbose', count=True)
def get_database(database: str, connection: str, verbose: bool):
    """Generate a network for a given database."""
    if verbose == 1:
        logging.basicConfig(level=logging.INFO)
    elif verbose == 2:
        logging.basicConfig(level=logging.DEBUG)

    from .database import install_and_populate_database

    click.secho(f'{EMOJI} Exporting {database}')
    install_and_populate_database(database, connection)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""This module has methods to acquire the networks for each database available in DiffuPath."""
import importlib
import logging
import os
import sys
from contextlib import redirect_stdout
from typing import Optional

import click
from bio2bel import AbstractManager
from biokeen.convert import to_pykeen_df, to_pykeen_path, to_pykeen_summary_path
from pybel import from_json_path, to_json_path

from .constants import EMOJI, OUTPUT_DIR

_PATHME_MODULES = {
    'kegg': 'pathme',
    'reactome': 'pathme',
    'wikipathways': 'pathme',
}

logger = logging.getLogger(__name__)


def install_and_populate_database(name: str, connection: Optional[str] = None) -> Optional[str]:
    """Install biological database.
    :param name: The name of the databae
    :param connection: The optional database connection
    """
    export_file = os.path.join(OUTPUT_DIR, f'{name}.csv')
    summary_file = os.path.join(OUTPUT_DIR, f'{name}_summary.csv')
    json_file = os.path.join(OUTPUT_DIR, f'{name}.bel.json')

    if os.path.exists(export_file):
        logger.info(f'{EMOJI} {name} has already been retrieved. See: {export_file}')
        return export_file

    if os.path.exists(json_file):
        logger.info(f'{EMOJI} loaded {name} JSON: {json_file}')
        graph = from_json_path(json_file)
        df = to_pykeen_df(graph)
        to_pykeen_path(df, export_file)
        to_pykeen_summary_path(df, summary_file)
        return export_file

    if name in _PATHME_MODULES:
        click.secho(
            f'{EMOJI} You are trying to install {name} which is part of PathMe. '
            'Due to the complexity of the installation, we refer to the tutorial at'
            ' https://github.com/PathwayMerger/PathMe to install it. Alternatively, you can download the dumps of '
            f'{name} directly from the URLs at the DiffuPath README file.'
        )
        sys.exit(0)

    module_name = f'bio2bel_{name}'

    bio2bel_module = ensure_bio2bel_installation(module_name)
    logger.debug(f'{EMOJI} imported {name}')

    manager_cls = bio2bel_module.Manager

    manager = manager_cls(connection=connection)

    if issubclass(manager_cls, AbstractManager):
        if not manager.is_populated():
            logger.info(f'{EMOJI} populating {module_name}')
            manager.populate()
        else:
            logger.debug(f'{EMOJI} {module_name} has already been populated')

    logger.debug(f'{EMOJI} generating BEL for {module_name}')
    graph = manager.to_bel()

    logger.debug(f'Summary: {graph.number_of_nodes()} nodes / {graph.number_of_edges()} edges')
    to_json_path(graph, json_file, indent=2)

    logger.debug(f'{EMOJI} generating PyKEEN TSV for {module_name}')
    df = to_pykeen_df(graph)
    to_pykeen_summary_path(df, summary_file)
    success = to_pykeen_path(df, export_file)

    if success:
        logger.debug(f'{EMOJI} wrote PyKEEN TSV to {export_file}')
        return export_file

    logger.warning(f'{EMOJI} no statements generated')


def ensure_bio2bel_installation(package: str):
    """Import a package, or install it."""
    try:
        b_module = importlib.import_module(package)

    except ImportError:
        logger.info(f'{EMOJI} pip install {package}')

        with redirect_stdout(sys.stderr):
            import subprocess
            subprocess.check_call([sys.executable, "-m", "pip", "install", package])

        try:
            return importlib.import_module(package)
        except ImportError:
            logger.exception(
                f'{EMOJI} failed to import {package}'
                f'{EMOJI} could not find {package} on PyPI. Try installing from GitHub with:'
                f'\n   pip install git+https://github.com/bio2bel/{package.split("_")[-1]}.git\n'
            )
            sys.exit(1)

    return b_module



# -*- coding: utf-8 -*-

"""Validate label_input utilities."""

import logging

log = logging.getLogger(__name__)


def validation_input_from_dict(mapping_by_subsets, diffuse_input_type, validation_type, input_diffuse):
    """Validate cross validation from dict."""
    input_labels = mapping_by_subsets[diffuse_input_type]
    validation_labels = mapping_by_subsets[validation_type]
    out_labels = set(input_diffuse.rows_labels) - input_labels - validation_labels

    return [input_labels, validation_labels, out_labels]


def validate_cross_validation_input_1(input_diffuse, input_validate, sets):
    """Validate cross validation input 1."""
    labeled, unlabeled, no_labeled = set(), set(), set()

    for score, i, j, row_label, col_label in input_diffuse.__iter__(get_indices=True, get_labels=True):
        if score == 1:
            labeled.add(row_label)

        if score == 0:
            unlabeled.add(row_label)

        if score == -1:
            no_labeled.add(row_label)

    assert labeled == sets[0]
    assert unlabeled == sets[1]
    assert no_labeled == sets[2]

    labeled, unlabeled, no_labeled = set(), set(), set()

    for score, i, j, row_label, col_label in input_validate.__iter__(get_indices=True, get_labels=True):
        if score == 1:
            labeled.add(row_label)

        if score == 0:
            unlabeled.add(row_label)

        if score == -1:
            no_labeled.add(row_label)

    assert labeled == sets[1]
    assert len(unlabeled) == 0
    assert no_labeled == sets[0].union(sets[2])

    logging.info('Test cross_validation_input_1 passed.')



# -*- coding: utf-8 -*-

"""Input mapping."""

import logging
from typing import Set, Tuple

import networkx as nx
from diffupy.kernels import regularised_laplacian_kernel
from diffupy.matrix import Matrix
from diffupy.process_input import generate_categoric_input_from_labels
from diffupy.utils import process_kernel_from_cli, process_network_from_cli, print_dict_dimensions

from .constants import EMOJI
from .utils import get_labels_set_from_dict, check_substrings

logger = logging.getLogger(__name__)


def get_mapping(
    to_map: Set,
    background_map: Set,
    mirnas=None,
    mirnas_mapping=None,
    submapping=None,
    title='',
    print_percentage=False,
):
    intersection = to_map.intersection(background_map)

    if mirnas and title in ['micrornas', 'Micrornas', 'mirna_nodes']:
        mirnas_substring = [e for e in check_substrings(mirnas, background_map) if 'mir' in e]
        intersection = intersection.union(mirnas_substring)

    if mirnas_mapping:
        intersection = intersection.union(mirnas_mapping.intersection(to_map))

    if submapping:
        intersection = intersection.intersection(submapping)

    if len(intersection) != 0 and print_percentage:
        print(f'{title} ({str(len(intersection))}) {(len(intersection) / len(to_map)) * 100}%')

    return intersection


def get_mapping_subsets(
    subsets_dict,
    map_labels,
    title,
    percentage_reference_labels=False,
    submapping=None,
    mirnas=None,
    mirnas_mapping=None,
    relative_statistics=None,
):
    entity_type_map = {'metabolite_nodes': 'metabolite', 'mirna_nodes': 'micrornas', 'gene_nodes': 'genes',
                       'bp_nodes': 'bps'}
    mapping_dict = {}
    total_entites = set()

    print('Mapping by ' + title + ':')

    if not isinstance(map_labels, set) and not isinstance(map_labels, list):
        raise Exception('map_labels must be a set or a list.')

    for type_name, entites in subsets_dict.items():

        # TODO: Mapping substring not other than mirnas

        mapping = get_mapping(entites,
                              map_labels,
                              mirnas=mirnas,
                              mirnas_mapping=mirnas_mapping,
                              title=type_name,
                              submapping=submapping)

        percentage = 0

        if percentage_reference_labels:
            percentage = len(mapping) / len(map_labels)
        elif relative_statistics:
            subset_len = len(relative_statistics[entity_type_map[type_name]])
            if subset_len != 0:
                percentage = len(mapping) / subset_len
        else:
            if len(entites) != 0:
                percentage = len(mapping) / len(entites)

        print(f'{type_name} ({str(len(mapping))}) {percentage * 100}%')

        mapping_dict[type_name] = (mapping, percentage)

        total_entites.update(mapping)

    total_dimention = len(total_entites)

    if percentage_reference_labels:
        percentage = total_dimention / len(map_labels)

    else:
        percentage = total_dimention / len(get_labels_set_from_dict(subsets_dict))

    print(f'Total ({total_dimention}) {percentage * 100}% \n')

    return mapping_dict, percentage, total_entites


def get_mapping_two_dim_subsets(
    two_dimentional_dict,
    map_labels,
    background_labels=None,
    percentage_reference_background_labels=False,
    mirnas=None,
    relative_statistics=None,
    mirnas_mapping=None,
):
    mapping_dict = {}
    total_entites = set()

    for subset_title, subsets_dict in two_dimentional_dict.items():
        mapping, percentage, entites = get_mapping_subsets(
            subsets_dict,
            map_labels,
            relative_statistics=relative_statistics,
            submapping=background_labels,
            mirnas=mirnas,
            mirnas_mapping=mirnas_mapping,
            title=subset_title.capitalize(),
        )

        mapping_dict[subset_title] = (mapping, percentage)
        total_entites.update(entites)

    total_dimention = len(total_entites)

    if percentage_reference_background_labels:
        total_percentage = total_dimention / len(background_labels)
    else:
        total_percentage = total_dimention / len(map_labels)

    print(f'Total ({total_dimention}) {total_percentage * 100}% \n')

    return mapping_dict, total_percentage, total_dimention


def process_input_from_cli(
    parse_set_funct,
    network_path,
    input_path,
    graph: bool = False,
    quantitative: bool = False,
) -> Tuple[Matrix, Matrix, set, nx.Graph]:
    """Process input from cli."""
    logger.info(f'{EMOJI} Loading networs (as graph or kernel) from {network_path} {EMOJI}')

    # Load label_input from dataset
    # TODO: Consider universal label_input processing, now parse set specific function by parameter, use from diffuPy process_input
    # input_scores = _process_input(input_path)
    dataset_labels_by_omics = parse_set_funct(input_path)
    dataset_all_labels = get_labels_set_from_dict(dataset_labels_by_omics)

    print_dict_dimensions(dataset_labels_by_omics, 'Dataset imported labels:')

    mirnas_dataset = dataset_labels_by_omics['micrornas']

    # Load background network from the label_input graph (if indicated as flag) or kernel
    if graph:
        graph = process_network_from_cli(network_path)

        logger.info(
            f'{EMOJI} Graph loaded with: \n'
            f'{graph.number_of_nodes()} nodes\n'
            f'{graph.number_of_edges()} edges\n'
            f'{EMOJI}'
        )

        # Generate the kernel from the label_input graph
        kernel = regularised_laplacian_kernel(graph)

    else:
        kernel = process_kernel_from_cli(network_path)

        logger.info(
            f'{EMOJI} Kernel loaded with: \n'
            f'{len(kernel.rows_labels)} nodes\n'
            f'{EMOJI}'
        )

    background_labels = kernel.rows_labels

    # Dataset label mapping to the network
    mapping_scores = get_mapping(
        dataset_all_labels,
        background_labels,
        title='Global mapping: ',
        mirnas=mirnas_dataset,
        print_percentage=True,
    )

    # Format label_input as Matrix for run_diffusion
    if quantitative:
        # TODO: Import from label_input column continuous scores as in diffuPy process_input
        # Generate label_input as a categoric label_input from labels
        input_scores = generate_categoric_input_from_labels(
            mapping_scores,
            'label_input with hidden true positives',
            kernel,
        )
    else:
        input_scores = generate_categoric_input_from_labels(
            mapping_scores,
            'label_input with hidden true positives',
            kernel,
        )

    return input_scores, kernel, mapping_scores, graph



# -*- coding: utf-8 -*-

import os
import re
import sys

sys.path.insert(0, os.path.abspath('../../src'))

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
    'sphinx_click.ext',
]
templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
project = 'diffupath'
copyright = '2020,  Josep Marín-Llaó, Sergi Picart Armada, Daniel Domingo-Fernández'
author = ' Josep Marín-Llaó, Sergi Picart Armada, Daniel Domingo-Fernández'

release = '0.0.2'

parsed_version = re.match(
    '(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(?:-(?P<release>[0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*))?(?:\+(?P<build>[0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*))?',
    release
)
version = parsed_version.expand('\g<major>.\g<minor>.\g<patch>')

tags = set()

if parsed_version.group('release'):
    tags.add('prerelease')

language = None
exclude_patterns = []
pygments_style = 'sphinx'
todo_include_todos = True
html_theme = 'sphinx_rtd_theme'
html_static_path = []
htmlhelp_basename = 'diffupathdoc'
latex_elements = {}
latex_documents = [
    (master_doc, 'diffupath.tex', 'DiffuPy Documentation',
     ' Josep Marín-Llaó, Sergi Picart Armada, Daniel Domingo-Fernández', 'manual'),
]
man_pages = [
    (master_doc, 'diffupath', 'diffupath Documentation', [author], 1)
]
texinfo_documents = [
    (master_doc, 'diffupath', 'diffupath Documentation', author, 'diffupath'),
]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'networkx': ('https://networkx.github.io/', None),
    'sqlalchemy': ('https://docs.sqlalchemy.org/en/latest', None),
    'pybel': ('https://pybel.readthedocs.io/en/latest/', None),
}

autodoc_member_order = 'bysource'
autoclass_content = 'both'

if os.environ.get('READTHEDOCS'):
    tags.add('readthedocs')



# -*- coding: utf-8 -*-

"""Tests validator module."""

import unittest

from diffupath.validate_input import validate_cross_validation_input_1

from diffupy.matrix import Matrix
from diffupy.process_input import generate_categoric_input_from_labels


class ValidateCVInputTest(unittest.TestCase):
    """Test cross validation."""

    def test_validate_cross_validation_input_1(self):
        """Test label_input of CV."""
        set_1 = {'adenine', 'adp', 'tyramine', 'nadph', 'glycine', 'glutathione', 'choline', 'inosine'}
        set_2 = {'g6pc3', 'oscar', 'lin37', 'pip4k2c', 'asap3', 'smg7', 'fosl1', 'atg3', 'plod3', 'apoe', 'calcoco2',
                 'hbegf', 'ccne1', 'magi3', 'gpc1', 'dusp5', 'idi1', 'ids', 'il12a', 'fancd2', 'zmpste24', 'znf383',
                 'gstm3', 'pgm2l1', 'dpm2', 'galk1', 'tfg', 'smox', 'znf473', 'thtpa', 'cyp2u1', 'rpa3', 'ggct', 'c5',
                 'foxa2', 'ndufb1', 'fn1', 'tyms', 'hsd17b2', 'birc5', 'tuba4a', 'ucn', 'dnm1l', 'dock4', 'actr3b',
                 'pigw', 'map2k1', 'tusc3', 'znf248', 'amdhd1', 'znf561', 'dck', 'lifr', 'gldc', 'rgl1', 'il1r1',
                 'actg1', 'agmat', 'ggt5', 'palb2', 'chac2', 'slc33a1', 'xylt2', 'il11', 'mir31', 'atm', 'serpind1',
                 'skp1', 'odc1', 'taldo1', 'mir331', 'gys1', 'cyp3a7', 'ctsk', 'man1a1', 'col2a1', 'arf6', 'acp2',
                 'mocos', 'cyp51a1', 'cbl', 'synj2'}
        set_3 = {'mir31', 'mir106b', 'mir193b', 'mir17', 'mir21', 'mir30e', 'mir30d'}

        all_labels = set_1 | set_2 | set_3

        background_matrix = Matrix(
            cols_labels=all_labels,
            rows_labels=all_labels,
            init_value=0,
            name='Test Matrix 1'
        )

        input_diffuse = generate_categoric_input_from_labels(
            set_1,
            'two out label_input',
            background_matrix,
            rows_unlabeled=set_2)

        input_validate = generate_categoric_input_from_labels(
            set_2,
            'two out label_input',
            background_matrix
        )

        validate_cross_validation_input_1(input_diffuse, input_validate, [set_1, set_2, set_3])



# -*- coding: utf-8 -*-

"""Visualization methods."""

import chart_studio.plotly as py
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objs as go
from matplotlib_venn import venn3


def heatmap(
        data, row_labels, col_labels, ax=None, cbar_kw=None, cbarlabel="", title="", **kwargs
):
    """Create a heatmap from a numpy array and two lists of labels.

    Optional parameters: ax: A matplotlib.axes.Axes instance to which the heatmap is plotted. If not provided, use
    current axes or create a new one. cbar_kw: A dictionary with arguments to :meth:`matplotlib.Figure.colorbar`.
    cbarlabel: The label for the colorbar.

    :param data: A 2D numpy array of shape (N,M)
    :param row_labels: A list or array of length N with the labels for the rows
    :param col_labels: A list or array of length N with the labels for the columns
    :param ax: axis
    :param cbar_kw: kwars for cbar
    :param cbarlabel: label
    :param title: title
    """

    if not ax:
        ax = plt.gca()

    # Plot the heatmap
    im = ax.imshow(data, **kwargs)

    # Create colorbar
    cbar = ax.figure.colorbar(im, ax=ax, **cbar_kw)
    cbar.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")

    # We want to show all ticks...
    ax.set_xticks(np.arange(data.shape[1]))
    ax.set_yticks(np.arange(data.shape[0]))
    # ... and label them with the respective list entries.
    ax.set_xticklabels(col_labels)
    ax.set_yticklabels(row_labels)

    # Let the horizontal axes labeling appear on top.
    ax.tick_params(top=True, bottom=False,
                   labeltop=True, labelbottom=False)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=-30, ha="right",
             rotation_mode="anchor")

    # Turn spines off and create white grid.
    for edge, spine in ax.spines.items():
        spine.set_visible(False)

    ax.set_xticks(np.arange(data.shape[1] + 1) - .5, minor=True)
    ax.set_yticks(np.arange(data.shape[0] + 1) - .5, minor=True)
    ax.grid(which="minor", color="w", linestyle='-', linewidth=3)
    ax.tick_params(which="minor", bottom=False, left=False)

    ax.set_title(title)

    return im, cbar


def annotate_heatmap(
        im,
        data=None,
        valfmt="{x:.2f}",
        textcolors=["black", "white"],
        entity_count=None,
        threshold=None,
        **textkw
):
    """Annotate a heatmap. Further parameters can be passed as textkw.

    :param im: The AxesImage to be labeled.
    :param data: Data used to annotate. If None, the image's data is used.
    :param valfmt: The sep of the annotations inside the heatmap. This should either use the string sep method.
    :param textcolors: A list or array of two color specifications.
    :param threshold: Value in data units according to which the colors from textcolors are applied.
    """

    if not isinstance(data, (list, np.ndarray)):
        data = im.get_array()

    # Normalize the threshold to the images color range.
    if threshold is not None:
        threshold = im.norm(threshold)
    else:
        threshold = im.norm(data.max()) / 2.

    # Set default alignment to center, but allow it to be
    # overwritten by textkw.
    kw = dict(horizontalalignment="center",
              verticalalignment="center")
    kw.update(textkw)

    # Get the formatter in case a string is supplied
    if isinstance(valfmt, str):
        valfmt = matplotlib.ticker.StrMethodFormatter(valfmt)

    # Loop over the data and create a `Text` for each "pixel".
    # Change the text's color depending on the data.
    texts = []
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            kw.update(color=textcolors[im.norm(data[i, j]) > threshold])
            text = im.axes.text(j, i, valfmt(entity_count[i, j], None), **kw)
            texts.append(text)

    return texts


def show_heatmap(count, percentage, databases, entity_types):
    """Show heatmap"""
    fig, ax = plt.subplots(figsize=(15, 7))

    im, cbar = heatmap(percentage, databases, entity_types, ax=ax,
                       cmap="YlGn", cbarlabel="percentage [0-1]")
    texts = annotate_heatmap(im, entity_count=count, valfmt="{x:1} ")

    fig.tight_layout()
    ax.set_title("Dataset 1 DiffuPath Mapping")

    plt.show()


def box_plot_from_dict(d, title='Box plot', x_title='x', y_title='y'):
    """Plot boxplot."""
    data = [go.Box(
        y=v,
        name=k
    ) for k, v in d.items()
    ]

    layout = go.Layout(
        title=go.layout.Title(
            text=title,
            xref='paper'
        ),
        xaxis=go.layout.XAxis(
            title=go.layout.xaxis.Title(
                text=x_title
            )
        ),
        yaxis=go.layout.YAxis(
            title=go.layout.yaxis.Title(
                text=y_title
            )
        )
    )

    fig = go.Figure(data=data, layout=layout)

    return py.iplot(fig)


def box_plot_from_two_dimension_dict(d, y_label='y'):
    """Boxplot two dimensions."""
    x = [k2 for k1, v1 in d.items() for k2, v2 in v1.items() for i in range(len(v2))]

    color_palete = ['#FF4136', '#3d9970', '#344f9e', '#bfc464']
    color_dict = {}
    i = 0

    for k, v in d.items():
        color_dict[k] = color_palete[i]
        i += 1

    data = [go.Box(
        y=[i for k2, v2 in v1.items() for i in v2],
        x=x,
        name=k1,
        marker=dict(
            color=color_dict[k1]
        )
    ) for k1, v1 in d.items()]

    layout = go.Layout(
        yaxis=dict(
            title=y_label,
            zeroline=False
        ),
        boxmode='group'
    )
    fig = go.Figure(data=data, layout=layout)

    return py.iplot(fig)


def show_venn_diagram(intersections, set_labels=('KEGG', 'Reactome', 'WikiPathways')):
    """Show venn diagram."""
    intersections_len = [len(intersection) for name, intersection in intersections.items()]

    plt.figure(figsize=(17, 8))
    v = venn3(subsets=intersections_len, set_labels=set_labels)

    plt.show()



# -*- coding: utf-8 -*-

"""Utils for sanity check."""

from collections import defaultdict

from openpyxl import load_workbook


def munge_labels(label):
    """Munge labels."""
    remove_set = ['*', ' ', '|', '-', '"', "'", "\n", "↑", "↓"]
    split_set = ['/', ',']

    label = str(label).lower()

    for symb in remove_set:
        if symb in label:
            label = label.replace(symb, '')

    if '/' in label:
        label = tuple(set(label.split('/')))
        if len(label) == 1:
            label = label[0]

    if ',' in label:
        label = tuple(set(label.split(',')))
        if len(label) == 1:
            label = label[0]

    return label


def parse_set1(path):
    """Process dataset 1."""
    wb = load_workbook(filename=path)
    sheet_titles = []
    omics_labels = defaultdict(lambda: set())

    for sheet in wb:
        cell_value = sheet['A3'].value

        # if "Expression data (FC) of the differentially expressed" in sheet['A1'].value:
        #    sheet_title = sheet['A1'].value.split("Expression data (FC) of the differentially expressed ",1)[1]
        #    sheet_title = sheet_title.split(" of HepG2 cells after treatment with ")
        #    sheet_title[1] = sheet_title[1].replace(". Statistical significance (p value < 0.05) is indicated.", "").replace(" CsA for", "")
        #    sheet_titles.append(sheet_title)

        if cell_value and ("Significant " in cell_value or "Metabolite" == cell_value):
            if cell_value == "Metabolite":
                sheet_title = ("Metabolite", '3 µM', ' 24h or 72h')
                min_row = 3

            else:
                sheet_title = cell_value.split("Significant ", 1)[1]
                sheet_title = sheet_title.split(" CsA ")
                sheet_title.append(sheet_title[1].split(" ")[0] + ' h')
                sheet_title[1] = sheet_title[0].split(" ")[1] + ' µM'
                sheet_title[0] = sheet_title[0].split(" ")[0]
                min_row = 4

            for col in sheet.iter_cols(min_row=min_row):
                col_label = col[0].value
                sheet_omic = sheet_title[0]

                if col_label in ['MicroRNA', 'hgnc symbol', 'Metabolite']:
                    # for cell in col[1:]:
                    # if munge_labels(cell.value) == '':
                    # print(cell)
                    # print(cell.value)
                    omics_labels[sheet_omic.lower()].update(
                        munge_labels(cell.value) for cell in col[1:] if munge_labels(cell.value) != '')

            sheet_titles.append(sheet_title)

    return omics_labels


def parse_set2(path):
    """Process dataset 2."""
    wb = load_workbook(filename=path)
    omics_labels = defaultdict(lambda: set())

    for sheet in wb:
        cell_value = sheet['A1'].value

        if "metabolites" in cell_value:
            sheet_title = "metabolite"
            min_row = 2

        elif "genes" in cell_value:
            sheet_title = "genes"
            min_row = 2

        else:
            continue

        for col in sheet.iter_cols(min_row=min_row):
            col_label = col[0].value

            if col_label in ['Metabolites', 'Metabolites name', 'Genes', 'microRNA']:
                if col_label == 'microRNA':
                    col_label = 'micrornas'

                if col_label == 'Metabolites':
                    col_label = 'metabolite'

                omics_labels[col_label.lower()].update(
                    munge_labels(cell.value) for cell in col[1:] if munge_labels(cell.value) != '')

    return omics_labels


def parse_set3(path):
    """Process dataset 3."""
    wb = load_workbook(filename=path)
    omics_labels = defaultdict(lambda: set())

    for sheet in wb:
        cell_value = sheet['A1'].value

        # if "Expression data (FC) of the differentially expressed" in sheet['A1'].value:
        #    sheet_title = sheet['A1'].value.split("Expression data (FC) of the differentially expressed ",1)[1]
        #    sheet_title = sheet_title.split(" of HepG2 cells after treatment with ")
        #    sheet_title[1] = sheet_title[1].replace(". Statistical significance (p value < 0.05) is indicated.", "").replace(" CsA for", "")
        #    sheet_titles.append(sheet_title)

        if "compounds" in cell_value:
            sheet_title = "metabolite"
            min_row = 2

        elif "mRNAs" in cell_value:
            sheet_title = "genes"
            min_row = 2

        elif "miRNAs selected" in cell_value:
            sheet_title = "micrornas"
            min_row = 2

        elif "pathway" in cell_value:
            sheet_title = "pathway"
            min_row = 2

        else:
            continue

        for col in sheet.iter_cols(min_row=min_row):
            col_label = col[0].value

            if col_label in ['Name', 'Abbreviation', 'miRNA']:
                omics_labels[sheet_title].update(
                    munge_labels(cell.value) for cell in col[1:] if munge_labels(cell.value) != '')

    return omics_labels



# -*- coding: utf-8 -*-

"""Cross-validation utilities."""

from collections import defaultdict

import numpy as np
from diffupy.diffuse_raw import diffuse_raw
from diffupy.matrix import Matrix
from diffupy.process_input import generate_categoric_input_from_labels
from sklearn import metrics
from tqdm import tqdm

from typing import Union, Tuple

from .topological_analyses import generate_pagerank_baseline
from .utils import random_disjoint_intersection_three_subsets, hide_true_positives, split_random_two_subsets

"""Random cross validation_datasets"""


def get_random_cv_split_input_and_validation(input: Union[list, set],
                                             background_mat: Matrix
                                             ) -> Tuple[Matrix, Matrix]:
    """Get random CV split."""
    input_labels, validation_labels = split_random_two_subsets(input)

    return (
        generate_categoric_input_from_labels(
            input_labels,
            'label_input with hidden true positives',
            background_mat
        ),
        generate_categoric_input_from_labels(
            validation_labels,
            'original label_input labels',
            background_mat
        )
    )


def get_random_cv_inputs_from_subsets_same_diff_input(input_subsets: Union[list, set],
                                                      background_mat
                                                      ):
    """Get random CV label_input from subsets with different label_input."""
    input_labels = set()
    input_unlabeled = set()

    validation_mats_by_entity_type = defaultdict()

    for entity_type, input in input_subsets.items():
        hidden_input = hide_true_positives(input[0])
        validation_mats_by_entity_type[entity_type] = generate_categoric_input_from_labels(
            input[0],
            'Dataset 1 ' + str(
                entity_type),
            background_mat
        )
        input_unlabeled.update(set(input[0]))
        input_labels.update(set(hidden_input))

    input_mat = generate_categoric_input_from_labels(input_labels, 'Dataset1', background_mat, input_unlabeled)

    return input_mat, validation_mats_by_entity_type


# Partial cross validation_datasets

def get_one_x_in_cv_inputs_from_subsets(
        input_subsets,
        background_mat,
        one_in='Reactome',
        rows_unlabeled=False,
        missing_value=-1
):
    """Get one cross label_input from subsets."""
    input_dict = {}
    input_labels = input_subsets.pop(one_in)
    rows_unlabel = None

    for labels_type, validation_labels in input_subsets.items():
        if rows_unlabeled:
            rows_unlabel = validation_labels
            missing_value = -1

        input_dict[labels_type] = (
            generate_categoric_input_from_labels(
                input_labels,
                'two out label_input',
                background_mat,
                missing_value,
                rows_unlabeled=rows_unlabel
            ),
            generate_categoric_input_from_labels(
                validation_labels,
                'two out label_input',
                background_mat,
                missing_value,
            )
        )
    return input_dict


def get_metrics(validation_labels,
                scores
                ):
    """Return metrics."""
    validation_labels_vec = validation_labels.__copy__()

    return metrics.roc_auc_score(validation_labels.mat, scores.mat), metrics.average_precision_score(
        validation_labels.mat, scores.mat)


def cross_validation_by_subset_same_diff_input(mapping_by_subsets,
                                               kernel,
                                               k=3,
                                               z=True):
    """Cross validation helper."""
    auroc_metrics = defaultdict(list)
    auprc_metrics = defaultdict(list)

    for i in tqdm(range(k)):
        input_mat, validation_inputs_by_subsets = get_random_cv_inputs_from_subsets_same_diff_input(
            mapping_by_subsets,
            kernel,
        )

        scores = diffuse_raw(graph=None, scores=input_mat, k=kernel, z=z)

        for entity, validation_labels in validation_inputs_by_subsets.items():
            auroc, auprc = get_metrics(validation_labels, scores)
            auroc_metrics[entity].append(auroc)
            auprc_metrics[entity].append(auprc)

    return auroc_metrics, auprc_metrics


def cross_validation_one_x_in(mapping_by_subsets,
                              kernel,
                              k=1,
                              missing_value=-1,
                              disjoint=False,
                              rows_unlabeled=False,
                              z=False
                              ):
    """Cross validation one."""
    auroc_metrics = defaultdict(lambda: defaultdict(list))
    auprc_metrics = defaultdict(lambda: defaultdict(list))

    scores_dict = defaultdict(lambda: defaultdict(list))
    validation_dict = defaultdict(lambda: defaultdict(list))
    input_dict = defaultdict(lambda: defaultdict(list))

    if disjoint:
        mapping_by_subsets = random_disjoint_intersection_three_subsets(mapping_by_subsets)

    for i in tqdm(range(k)):

        for diffuse_input_type in tqdm(mapping_by_subsets):
            inputs = get_one_x_in_cv_inputs_from_subsets(
                dict(mapping_by_subsets),
                kernel,
                one_in=diffuse_input_type,
                rows_unlabeled=rows_unlabeled,
                missing_value=missing_value,
            )

            for validation_type, validation_labels in inputs.items():
                input_diffuse, input_validation = validation_labels[0], validation_labels[1]

                # Input test
                # validate_cross_validation_input_1(input_diffuse, input_validation, validation_input_from_dict(mapping_by_subsets, diffuse_input_type, validation_type, input_diffuse))

                # Run diffusion
                scores = diffuse_raw(graph=None, scores=input_diffuse, k=kernel, z=z)

                scores.cols_labels = ['scores']
                input_validation.cols_labels = ['input_validation']
                input_diffuse.cols_labels = ['input_diffuse']

                auroc, auprc = get_metrics(input_validation, scores)

                auroc_metrics[diffuse_input_type][validation_type].append(auroc)
                auprc_metrics[diffuse_input_type][validation_type].append(auprc)

                scores.col_bind(matrix=input_validation)
                scores.col_bind(matrix=input_diffuse)

                scores_dict[diffuse_input_type][validation_type].append(scores)

    return dict(auroc_metrics), dict(auprc_metrics), dict(scores_dict)


# Method cross validation_datasets


def generate_random_score_ranking(background_mat):
    """Generate random scores."""
    return Matrix(
        mat=np.random.rand(len(background_mat.rows_labels)),
        rows_labels=background_mat.rows_labels,
        cols_labels=['Radom_Baseline'],
    )


def cross_validation_by_method(all_labels_mapping,
                               graph,
                               kernel,
                               k=100
                               ):
    """Cross validation by method."""
    auroc_metrics = defaultdict(list)
    auprc_metrics = defaultdict(list)

    for _ in tqdm(range(k)):
        input_diff, validation_diff = get_random_cv_split_input_and_validation(
            all_labels_mapping, kernel
        )

        scores_z = diffuse_raw(graph=None, scores=input_diff, k=kernel, z=True)
        scores_raw = diffuse_raw(graph=None, scores=input_diff, k=kernel, z=False)
        scores_page_rank = generate_pagerank_baseline(graph, kernel)

        method_validation_inputs = {
            'raw': (validation_diff,
                    scores_raw
                    ),
            'z': (validation_diff,
                  scores_z
                  ),
            'random_baseline': (
                validation_diff,
                generate_random_score_ranking(kernel)
            ),
            'page_rank_baseline': (
                validation_diff,
                scores_page_rank
            ),

        }

        for method, validation_set in method_validation_inputs.items():
            auroc, auprc = get_metrics(*validation_set)
            auroc_metrics[method].append(auroc)
            auprc_metrics[method].append(auprc)

    return auroc_metrics, auprc_metrics



# -*- coding: utf-8 -*-

"""Topological analyses."""

from collections import defaultdict

import warnings

import networkx as nx
import numpy as np
from diffupy.matrix import LaplacianMatrix, Matrix
from diffupy.utils import get_simple_graph_from_multigraph


def generate_pagerank_baseline(graph: nx.Graph, background_mat: Matrix) -> Matrix:
    """Generate baseline results using page rank algorithm."""
    graph = get_simple_graph_from_multigraph(graph)

    pagerank_scores = nx.pagerank(graph)

    if len(pagerank_scores.values()) != len(background_mat.mat):
        warnings.warn('The provided graph do not match the kernel nodes amount. '
                       'The nodes will be matched (deleting and filling missing) according to the reference Matrix.')

    return Matrix(mat=np.array(
        list(pagerank_scores.values())).reshape(
        (len(list(pagerank_scores.values())), 1)
    ),
        rows_labels=list(pagerank_scores.keys()),
        cols_labels=['PageRank']
    ).match_delete_rows(background_mat.rows_labels).match_missing_rows(background_mat.rows_labels).match_rows(background_mat)


def resistance_distance(G=None, M=None, normalized=False):
    """Calculate the resistance."""
    if G:
        ER = LaplacianMatrix(G, normalized)
        add_edges_inv = 1 / G.number_of_edges()

    elif M:
        ER = M
        add_edges_inv = 1 / len(ER.rows_labels)

    else:
        raise Warning('A graph or a matrix must be given.')

    LT = np.linalg.inv([x + add_edges_inv
                        for x in ER.mat
                        ]
                       )
    dLT = np.diag(LT)

    sER = [x * -2
           for x in LT
           ] + dLT[:, np.newaxis]
    ssER = sER + dLT[np.newaxis, :]

    ER.mat = ssER

    return ER


def filter_quadratic_mat_by_mapping(M, mapping):
    d = defaultdict(lambda: defaultdict(lambda: list()))

    for k1, v1 in mapping.items():
        for k2, v2 in mapping.items():
            for e1 in v1[0]:
                for e2 in v2[0]:
                    d[k1][k2].append(M.get_cell_from_labels(e1, e2))

    return d
